// ASIOHostDemo.cpp : Defines the entry point for the application.
//
#include "stdafx.h"
#include <Windows.h>
#include <stdio.h>
#include <math.h>
#include "ASIOHost.h"
#include "resource.h"

#define TWO_PI (6.28318530717959) // PI * 2
#define Max16BitAmplitude (0x8000 >> 1)  // 16 bit audio down 6 dB

#define STEREO_CHANNELS_FOR_TESTING 1
#define MONO_CHANNELS_FOR_TESTING 2
#define MAX_LOADSTRING 100
#define MAXPATH 256
#define SCREEN_WIDTH 400
#define SCREEN_HEIGHT 200

// Global Variables:
HINSTANCE hInst;
HWND hwndApp;
TCHAR szTitle[MAX_LOADSTRING];
TCHAR szWindowClass[MAX_LOADSTRING];
char pPathname[MAXPATH];
char pCurrentDirPath[MAXPATH];

MODE modeSelected = MODE_NOT_SET;

double dLPhaseAngle = 0;
double dRPhaseAngle = 0;
double dPhaseAngle = 0;
double dRadians = 0;
DWORD dwBufferSampleSize = 0;

int nMaxChannelsAvailable = 0;
int nChannelsRequested = 0;
int nChosenDevice = -1;

short *pInterleaveBuf = NULL; // for stereo record

// from ASIOHost.h
ChannelVars LeftChannelVars; 
ChannelVars RightChannelVars;
ChannelVars StereoChannelVars;

short PlayBytesPerSample = 2;
short RecordBytesPerSample = 2;

FILE *fp = NULL;
FILE *fpLeft = NULL;
FILE *fpRight = NULL;

bool bASIOInitialized = false;
bool bSetChannelMode = false;
bool bMonoMode = false;
bool bRecording = false;
bool bCreatedStereoVirtualChn = false;
bool bCreatedLeftVirtualChn = false;
bool bCreatedRightVirtualChn = false;

ATOM				MyRegisterClass(HINSTANCE hInstance);
BOOL				InitInstance(HINSTANCE, int);
LRESULT CALLBACK	WndProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK	About(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK    SelectDevice(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
void CalculateBufferSize(void);
void PrepareVirtualChannels(void);
void PrepareFileIO(void);
void InitSinewaveGenerator(void);
void GetMonoSamples(DWORD dwSamples, LPBYTE pSndBuf, double *pPhaseAngle);
void GetStereoSamples(DWORD dwSamples, LPBYTE pLeftBuf, LPBYTE pRightBuf);
void OnLeftClickMono(HWND hWnd, LPARAM lParam);
void OnLeftClickStereo(HWND hWnd, LPARAM lParam);


void SetChannelMode(MODE mode)
{
 if(bSetChannelMode)
 {
    MessageBox(NULL, TEXT("You already selected the mode!"), TEXT("ASIOHostDemo..."), MB_OK);
	return;
 }
 if(nChosenDevice == -1)
 {
    MessageBox(NULL, TEXT("You need to select the device first!"), TEXT("ASIOHostDemo..."), MB_OK);
	return;
 }
 if(!bASIOInitialized)
 {
    MessageBox(NULL, TEXT("You need to select the device first and initialize ASIO first!"), TEXT("ASIOHostDemo..."), MB_OK);
	return;
 }
 if((mode == MONO_PLAY_ONLY) || (mode == MONO_RECORD_ONLY) || (mode == MONO_PLAY_RECORD))
 {
	 bMonoMode = true;
 }


 if((mode == MONO_RECORD_ONLY) || (mode == MONO_PLAY_RECORD) || (mode == STEREO_RECORD_ONLY) || (mode == STEREO_PLAY_RECORD))
 {
	 bRecording = true;
 }

 nMaxChannelsAvailable = ASIOGetMaxVirtualChannels(mode);
 if(!nMaxChannelsAvailable)return;

 // Before we create WavBuf we must set the maximum number of channels we plan to use,
 // and of course we must ensure nChannelsRequested <= nMaxChannelsAvailable
 // Though we are not obligated to use them all in the end,
 // creating more channels than we need is a waste of resources.
 if(bMonoMode)
 {
    nChannelsRequested = MONO_CHANNELS_FOR_TESTING;
 }else
 {
    nChannelsRequested = STEREO_CHANNELS_FOR_TESTING;
 }

 if(!nChannelsRequested || (nChannelsRequested > nMaxChannelsAvailable))
 {
	return;
 }

 // This is a key final initialzation step. 
 if(nChannelsRequested != ASIOCreateBuffers(nChannelsRequested, mode))
 {
    return;
 } 

 modeSelected = mode;
 bSetChannelMode = true;

 // Set default values for 16 bit samples...
 PlayBytesPerSample = 2;
 RecordBytesPerSample = 2;
 CalculateBufferSize();
 InitSinewaveGenerator();
 PrepareVirtualChannels();
 if(bRecording)PrepareFileIO();
 if(bRecording && !bMonoMode)
    pInterleaveBuf = new short [dwBufferSampleSize * 2];
}

void CalculateBufferSize(void)
{
 // Calculate how many samples in 100 milliseconds at current sample rate...
 dwBufferSampleSize = ASIOGetSampleRate() / 10;
 // Now ensure we have an exact multiple of the current primary buffer size...
 DWORD dwPrimaryBufferSampleSize = ASIOGetPrimaryBufferSizeSamples();
 DWORD n = dwBufferSampleSize / dwPrimaryBufferSampleSize;
 if((dwBufferSampleSize % dwPrimaryBufferSampleSize) != 0)++n;
 if(n < 2)n = 2;
 dwBufferSampleSize = n * dwPrimaryBufferSampleSize;
}

void RefillLeftBuffer(LPBYTE pSndBuf, DWORD dwBufsizeBytes, DWORD dwInstanceData)
{
 GetMonoSamples(dwBufsizeBytes / PlayBytesPerSample, pSndBuf, &dLPhaseAngle);
}

void RefillRightBuffer(LPBYTE pSndBuf, DWORD dwBufsizeBytes, DWORD dwInstanceData)
{
 GetMonoSamples(dwBufsizeBytes / PlayBytesPerSample, pSndBuf, &dRPhaseAngle);
}

void RefillStereoBuffer(LPBYTE pSndBuf, DWORD dwBufsizeBytes, DWORD dwInstanceData)
{
 GetStereoSamples(dwBufsizeBytes / PlayBytesPerSample, pSndBuf, (pSndBuf + dwBufsizeBytes));
}

void SaveLeftBuffer(LPBYTE pSndBuf, DWORD dwBufsizeBytes, DWORD dwInstanceData)
{
 fwrite(pSndBuf, dwBufsizeBytes, 1, fpLeft);
}

void SaveRightBuffer(LPBYTE pSndBuf, DWORD dwBufsizeBytes, DWORD dwInstanceData)
{
 fwrite(pSndBuf, dwBufsizeBytes, 1, fpRight);
}

void SaveStereoBuffer(LPBYTE pSndBuf, DWORD dwBufsizeBytes, DWORD dwInstanceData)
{
 short *pLeftChn = (short*)pSndBuf;
 short *pRightChn = (short*)(pSndBuf+dwBufsizeBytes);
 int samples = dwBufsizeBytes / RecordBytesPerSample;
 int i, j = 0;
 for(i = 0; i < samples; i++)
 {
     pInterleaveBuf[j++] = pLeftChn[i];
	 pInterleaveBuf[j++] = pRightChn[i];
 }
 fwrite(pInterleaveBuf, sizeof(short), j, fp);
}

void PrepareVirtualChannels(void)
{
 if(bMonoMode)
 {
    ZeroMemory(&LeftChannelVars, sizeof(ChannelVars));
    LeftChannelVars.nVirtualChnNum = 0;
    LeftChannelVars.nBufSizeSamples = dwBufferSampleSize;
    LeftChannelVars.pOutBufServiceProc = RefillLeftBuffer;
    LeftChannelVars.pInBufServiceProc = SaveLeftBuffer;
    bCreatedLeftVirtualChn = CreateVirtualChannel(&LeftChannelVars);

    ZeroMemory(&RightChannelVars, sizeof(ChannelVars));
    RightChannelVars.nVirtualChnNum = 1;
    RightChannelVars.nBufSizeSamples = dwBufferSampleSize;
	RightChannelVars.pOutBufServiceProc = RefillRightBuffer;
	RightChannelVars.pInBufServiceProc = SaveRightBuffer;
    bCreatedRightVirtualChn = CreateVirtualChannel(&RightChannelVars);
 }else // stereo 
 { 
    ZeroMemory(&StereoChannelVars, sizeof(ChannelVars));
    StereoChannelVars.nVirtualChnNum = 0;
    StereoChannelVars.nBufSizeSamples = dwBufferSampleSize;
    StereoChannelVars.pOutBufServiceProc = RefillStereoBuffer;
    StereoChannelVars.pInBufServiceProc = SaveStereoBuffer;
    bCreatedStereoVirtualChn = CreateVirtualChannel(&StereoChannelVars);
 }
}

void PrepareFileIO(void)
{
 if(bMonoMode)
 {
    strcpy(pPathname, pCurrentDirPath);
    strcat(pPathname, "\\left.pcm");
    fpLeft = fopen(pPathname, "wb");
    strcpy(pPathname, pCurrentDirPath);
    strcat(pPathname, "\\right.pcm");
    fpRight = fopen(pPathname, "wb");
 }else // stereo 
 { 
    strcpy(pPathname, pCurrentDirPath);
    strcat(pPathname, "\\stereo.pcm");
    fp = fopen(pPathname, "wb");
 }
}


int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
	MSG msg;
	HACCEL hAccelTable;

	// Initialize global strings
	LoadString(hInstance, IDS_APP_TITLE, szTitle, MAX_LOADSTRING);
	LoadString(hInstance, IDC_ASIOHOSTDEMO, szWindowClass, MAX_LOADSTRING);
	MyRegisterClass(hInstance);

	// Perform application initialization:
	if (!InitInstance (hInstance, nCmdShow)) 
	{
		return FALSE;
	}

	hAccelTable = LoadAccelerators(hInstance, (LPCTSTR)IDC_ASIOHOSTDEMO);

	// Main message loop:
	while (GetMessage(&msg, NULL, 0, 0)) 
	{
		if (!TranslateAccelerator(msg.hwnd, hAccelTable, &msg)) 
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	return msg.wParam;
}


ATOM MyRegisterClass(HINSTANCE hInstance)
{
	WNDCLASSEX wcex;

	wcex.cbSize = sizeof(WNDCLASSEX); 

	wcex.style			= CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc	= (WNDPROC)WndProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= LoadIcon(hInstance, (LPCTSTR)IDI_ASIOHOSTDEMO);
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
	wcex.lpszMenuName	= MAKEINTRESOURCE(IDC_ASIOHOSTDEMO);
	wcex.lpszClassName	= szWindowClass;
	wcex.hIconSm		= LoadIcon(wcex.hInstance, (LPCTSTR)IDI_SMALL);

	return RegisterClassEx(&wcex);
}

BOOL InitInstance(HINSTANCE hInstance, int nCmdShow)
{
   HWND hWnd;

   hWnd = CreateWindow(szWindowClass, szTitle, WS_OVERLAPPEDWINDOW,
          CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL);

   if(!hWnd)
   {
      return FALSE;
   }
   hInst = hInstance;
   hwndApp = hWnd;
   ShowWindow(hWnd, nCmdShow);
   UpdateWindow(hWnd);

   return TRUE;
}

void CleanUp(void)
{
  if(bCreatedStereoVirtualChn != NULL)
     CloseVirtualChannel(0);
  if(bCreatedLeftVirtualChn != NULL)
     CloseVirtualChannel(0);
  if(bCreatedRightVirtualChn != NULL)
     CloseVirtualChannel(1);

  ASIOShutdown();

  if(pInterleaveBuf != NULL)
     delete [] pInterleaveBuf;

  if(fp != NULL)
  {
     fclose(fp);
  }
  if(fpLeft != NULL)
  {
     fclose(fpLeft);
  }
  if(fpRight != NULL)
  {
     fclose(fpRight);
  }
}


LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	int wmId, wmEvent;
	PAINTSTRUCT ps;
	HDC hdc;

	switch (message) 
	{
		case WM_CREATE:
         ZeroMemory(pCurrentDirPath, MAXPATH);
		 GetCurrentDirectoryA(MAXPATH, pCurrentDirPath);
		 {
          int xborder = 2 * GetSystemMetrics(SM_CXFRAME);
          int yborder = 2 * GetSystemMetrics(SM_CYFRAME);
          int caption = GetSystemMetrics(SM_CYCAPTION);
          int menu = GetSystemMetrics(SM_CYMENU);
          SetWindowPos(hWnd, NULL, 0, 0, SCREEN_WIDTH + xborder, SCREEN_HEIGHT + yborder + caption + menu, SWP_NOCOPYBITS | SWP_NOMOVE | SWP_NOREPOSITION | SWP_NOZORDER);
		 }
 		break;
		case WM_LBUTTONDOWN:
			if(bMonoMode)
			{
			   OnLeftClickMono(hWnd, lParam);
			}else
			{
               OnLeftClickStereo(hWnd, lParam);
			}
		break;
		case WM_WANTS_RESET:
         // A device may request a reset at any time.
         // Some sound cards require a complete reset when an application requests a change in sample rate
         // In this case, ASIOHost will send a WM_WANTS_RESET message, as defined in constants.h
         MessageBox(NULL, TEXT("Please exit the application and restart it for change to take effect.\0"),
                    TEXT("Sample rate change request...\0"), MB_OK | MB_ICONEXCLAMATION);
		break;
		case WM_COMMAND:
			wmId    = LOWORD(wParam); 
			wmEvent = HIWORD(wParam); 
			// Parse the menu selections:
			switch (wmId)
			{
				case IDM_ABOUT:
				   DialogBox(hInst, (LPCTSTR)IDD_ABOUTBOX, hWnd, (DLGPROC)About);
				   break;
				case ID_VIEWDEVICES_SELECTDEVICE:
				   if((nChosenDevice >= 0) || bASIOInitialized)
				   {
		              MessageBox(NULL, TEXT("You already selected an ASIO device!"), TEXT("ASIO Devices..."), MB_OK);
				      break;
				   }
				   DialogBox(hInst, (LPCTSTR)IDD_DEVICES, hWnd, (DLGPROC)SelectDevice);
				   break;

                // I need to grey out these menus once the mode is set!!!
				case ID_SETMODE_MONO_PLAY:
				   SetChannelMode(MONO_PLAY_ONLY);
				   break;
				case ID_SETMODE_MONO_RECORD:
				   SetChannelMode(MONO_RECORD_ONLY);
				   break;
				case ID_SETMODE_MONO_BOTH:
				   SetChannelMode(MONO_PLAY_RECORD);
				   break;
				case ID_SETMODE_STEREO_PLAY:
				   SetChannelMode(STEREO_PLAY_ONLY);
				   break;
				case ID_SETMODE_STEREO_RECORD:
				   SetChannelMode(STEREO_RECORD_ONLY);
				   break;
				case ID_SETMODE_STEREO_BOTH:
				   SetChannelMode(STEREO_PLAY_RECORD);
				   break;

				case IDM_FILE_START:
                 if(nChosenDevice == -1)
                 {
                    MessageBox(NULL, TEXT("You need to select the device first!"), TEXT("ASIOHostDemo..."), MB_OK);
	                break;
                 }
				 if(!bSetChannelMode)
                 {
                    MessageBox(NULL, TEXT("You need to select the mode first!"), TEXT("ASIOHostDemo..."), MB_OK);
	                break;
                 }

                 if(bMonoMode)
				 {
                    if(bCreatedLeftVirtualChn && bCreatedRightVirtualChn)
					{
                       PreLoadBuffers(0); PreLoadBuffers(1);
                       StartPlay(0); StartPlay(1);
					}
				 }else
				 {
                    if(bCreatedStereoVirtualChn)
					{
                       PreLoadBuffers(0);
           	           StartPlay(0);
					}
				 }
				 break;

				case IDM_FILE_STOP:
                 if(nChosenDevice == -1)
                 {
                    MessageBox(NULL, TEXT("You need to select the device first!"), TEXT("ASIOHostDemo..."), MB_OK);
	                break;
                 }
				 if(!bSetChannelMode)
                 {
                    MessageBox(NULL, TEXT("You need to select the mode first!"), TEXT("ASIOHostDemo..."), MB_OK);
	                break;
                 }
                 if(bMonoMode)
				 {
                    StopPlay(0);
                    StopPlay(1);
				 }else
				 {
           	        StopPlay(0);
				 }
				 break;

				case IDM_EXIT:
				 DestroyWindow(hWnd);
				   break;
				default:
				   return DefWindowProc(hWnd, message, wParam, lParam);
			}
			break;
		case WM_PAINT:
			hdc = BeginPaint(hWnd, &ps);
			EndPaint(hWnd, &ps);
			break;
		case WM_DESTROY:
			CleanUp();
			PostQuitMessage(0);
			break;
		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
   }
   return 0;
}

LRESULT CALLBACK About(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
		case WM_INITDIALOG:
				return TRUE;

		case WM_COMMAND:
			if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) 
			{
				EndDialog(hDlg, LOWORD(wParam));
				return TRUE;
			}
			break;
	}
    return FALSE;
}

// *Note - devices combo must not sort!
LRESULT CALLBACK SelectDevice(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
 HWND hWndCombo;
 char *pStrBuf, *ptr;
 int n, nDevices;

 switch (message)
 {
  case WM_INITDIALOG:
   pStrBuf = new char [2048];
   if(pStrBuf == NULL)return TRUE;
   ZeroMemory(pStrBuf, 2048);
   nDevices = ASIOGetDevices(pStrBuf, 2048);
   if(!nDevices)
   {
      delete [] pStrBuf;
	  return TRUE;
   }
   hWndCombo = GetDlgItem(hDlg, IDC_COMBO1);
   ptr = pStrBuf;
   for(n = 0; n < nDevices; n++)
   {
	   SendMessage(hWndCombo, CB_ADDSTRING, 0, (LPARAM) (LPCTSTR)ptr);
       // Device strings are separated by a NULL. Skip to next...
	   ptr += strlen(ptr);
	   ++ptr; 
   }
  delete [] pStrBuf;
  return TRUE;

  case WM_COMMAND:
   if(LOWORD(wParam) == IDOK) 
   {
      hWndCombo = GetDlgItem(hDlg, IDC_COMBO1);
	  nChosenDevice = SendMessage(hWndCombo, CB_GETCURSEL, 0, 0);
	  if(nChosenDevice >= 0)
	  {
		 // Now we have the desired device, go ahead and initialize ASIO...
		 bASIOInitialized = ASIOInit(hwndApp, nChosenDevice);
	  }
	  EndDialog(hDlg, LOWORD(wParam));
	  return TRUE;
   }
   if(LOWORD(wParam) == IDCANCEL) 
   {
	  EndDialog(hDlg, LOWORD(wParam));
	  return TRUE;
   }
  break;
 }
 return FALSE;
}

void InitSinewaveGenerator(void)
{
 double dFrequency;
 DWORD dwSampleRate = ASIOGetSampleRate();
 dFrequency = ((double)dwSampleRate / 100.0);
 // Each full sine wave will be exactly 100 samples,
 // dFrequency = 441 Hz @ sample rate 44,100
 // dFrequency = 480 Hz @ sample rate 48,000
 // dFrequency = 960 Hz @ sample rate 96,000
 // etc.
 dRadians = TWO_PI * dFrequency / (double)dwSampleRate;
 dPhaseAngle = 0;
 dLPhaseAngle = 0;
 dRPhaseAngle = 0;
 dPhaseAngle = 0;
}


void GetMonoSamples(DWORD dwSamples, LPBYTE pSndBuf, double *pPhaseAngle)
{
 short *pBuf = (short*)pSndBuf;
 double dMaxAmplitude = Max16BitAmplitude;

 for(DWORD i = 0; i < dwSamples; i++)
 {
     double dSample = dMaxAmplitude * sin(*pPhaseAngle);
	 if(dSample >= 0)
	 {
        *pBuf++ = (short)floor(dSample);
	 }else
	 {
        *pBuf++ = (short)ceil(dSample);
	 }
	 *pPhaseAngle += dRadians;
 }

 if(*pPhaseAngle > TWO_PI)
 {
    *pPhaseAngle = fmod(*pPhaseAngle, TWO_PI);
 }
}

void GetStereoSamples(DWORD dwSamples, LPBYTE pLeftBuf, LPBYTE pRightBuf)
{
 short *pLeftSndBuf = (short*)pLeftBuf;
 short *pRightSndBuf = (short*)pRightBuf;
 double dMaxAmplitude = Max16BitAmplitude;

 for(DWORD i = 0; i < dwSamples; i++)
 {
     double dSample = dMaxAmplitude * sin(dPhaseAngle);
	 if(dSample >= 0)
	 {
        *pLeftSndBuf++ = (short)floor(dSample);
        *pRightSndBuf++ = (short)floor(dSample);
	 }else
	 {
        *pLeftSndBuf++ = (short)ceil(dSample);
        *pRightSndBuf++ = (short)ceil(dSample);
	 }
	 dPhaseAngle += dRadians;
 }

 if(dPhaseAngle > TWO_PI)
 {
    dPhaseAngle = fmod(dPhaseAngle, TWO_PI);
 }
}


// If you left mouse click in the lower left corner of the application screen
// you can Pause, and then click in the lower right corner to resume.
// If you left mouse click in the central area of the application screen
// the higher up you click the greater you can reduce the volume,
// and then the lower down you click the greater you raise the volume back up.
// Click on the bottom of screem for full volume.
void OnLeftClickMono(HWND hWnd, LPARAM lParam)
{
 RECT rect;
 int xPos, yPos;
 if(bCreatedLeftVirtualChn && bCreatedRightVirtualChn)
 { 
	GetClientRect(hwndApp, &rect);
	yPos = HIWORD(lParam);
	if(yPos >= (rect.bottom - 18))
    {
	   xPos = LOWORD(lParam);
	   if(xPos < (rect.right / 4))
       {
	      Pause(0);
	      Pause(1);
       }else if(xPos > ((rect.right / 4) * 3))
       {
	      Resume(0);
	      Resume(1);
       }
       return;
    }
	if(yPos >= 160)yPos = 159;
	SetVolume(0, yPos/10);
	SetVolume(1, yPos/10);
 }
}

void OnLeftClickStereo(HWND hWnd, LPARAM lParam)
{
 RECT rect;
 int xPos, yPos;
 if(bCreatedStereoVirtualChn)
 { 
	GetClientRect(hwndApp, &rect);
	yPos = HIWORD(lParam);
	if(yPos >= (rect.bottom - 18))
    {
	   xPos = LOWORD(lParam);
	   if(xPos < (rect.right / 4))
       {
	      Pause(0);
       }else if(xPos > ((rect.right / 4) * 3))
       {
          Resume(0);
       }
	   return;
	}
	if(yPos >= 160)yPos = 159;
	SetVolume(0, yPos/10);
 }
}
